# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements functions that write files from matador
documents or Crystal objects.
"""

import os
from traceback import print_exc

import numpy as np
import pymongo as pm

from matador.utils.cell_utils import cart2abcstar, frac2cart, cart2abc
from matador.utils.cell_utils import abc2cart, calc_mp_grid
from matador.utils.cursor_utils import display_results
from matador.utils.chem_utils import get_formula_from_stoich, get_stoich, get_root_source
from matador.swaps import AtomicSwapper
from .utils import file_writer_function, generate_relevant_path, generate_hash

EPS = 1e-8


def query2files(
    cursor, dirname=None, max_files=10000, top=None, prefix=None,
    cell=None, param=None, res=None, pdb=None, json=None, xsf=None, markdown=True, latex=False,
    subcmd=None, argstr=None,
    **kwargs
):
    """ Many-to-many convenience function for many structures being written to
    many file types.

    Parameters:
        cursor (:obj:`list` of :obj:`dict`/:class:`AtomicSwapper`): list of matador dictionaries to write out.

    Keyword arguments:
        dirname (str): the folder to save the results into. Will be created if non-existent.
            Will have integer appended to it if already existing.
        max_files (int): if the number of files to be written exceeds this number, then raise RuntimeError.
        **kwargs (dict): dictionary of {filetype: bool(whether to write)}. Accepted file types
            are cell, param, res, pdb, json, xsf, markdown and latex.

    """
    multiple_files = any((cell, param, res, pdb, xsf))
    prefix = prefix + '-' if prefix is not None else ''

    if isinstance(cursor, AtomicSwapper):
        cursor = cursor.cursor
        subcmd = "swaps"

    if subcmd in ['polish', 'swaps']:
        info = False
        hash_dupe = False
    else:
        info = True
        hash_dupe = False

    if isinstance(cursor, list):
        num = len(cursor)
    else:
        num = cursor.count()

    if top is not None:
        if top < num:
            num = top

    num_files = num * sum(1 for ext in [cell, param, res, pdb, xsf] if ext)

    if multiple_files:
        print('Intending to write', num, 'structures to file...')
        if num_files > max_files:
            raise RuntimeError(
                "Not writing {} files as it exceeds argument `max_files` limit of {}"
                .format(num_files, max_files)
            )

    if dirname is None:
        dirname = generate_relevant_path(subcmd=subcmd, **kwargs)

    _dir = False
    dir_counter = 0
    # postfix integer on end of directory name if it exists
    while not _dir:
        if dir_counter != 0:
            directory = dirname + str(dir_counter)
        else:
            directory = dirname
        if not os.path.isdir(directory):
            os.makedirs(directory)
            _dir = True
        else:
            dir_counter += 1

    for _, doc in enumerate(cursor[:num]):
        # generate an appropriate filename for the structure
        root_source = get_root_source(doc)

        if '_swapped_stoichiometry' in doc:
            formula = get_formula_from_stoich(doc['_swapped_stoichiometry'])
        else:
            formula = get_formula_from_stoich(doc['stoichiometry'])

        if subcmd == 'swaps':
            root_source = root_source.replace('-swap-', '-')

        name = root_source

        if 'OQMD ' in root_source:
            name = '{formula}-OQMD_{src}'.format(formula=formula, src=root_source.split(' ')[-1])
        elif 'mp-' in root_source:
            name = '{formula}-MP_{src}'.format(formula=formula, src=root_source.split('-')[-1])
        if 'icsd' in doc and 'CollCode' not in name:
            name += '-CollCode{}'.format(doc['icsd'])
        else:
            pf_id = None
            for source in doc['source']:
                if 'pf-' in source:
                    pf_id = source.split('-')[-1]
                    break
            else:
                if 'pf_ids' in doc:
                    pf_id = doc['pf_ids'][0]
            if pf_id is not None:
                name += '-PF-{}'.format(pf_id)

        # if swaps, prepend new composition
        if subcmd == 'swaps':
            new_formula = get_formula_from_stoich(get_stoich(doc['atom_types']))
            name = '{}-swap-{}'.format(new_formula, name)

        path = "{directory}/{prefix}{name}".format(directory=directory, prefix=prefix, name=name)

        if param:
            doc2param(doc, path, hash_dupe=hash_dupe)
        if cell:
            doc2cell(doc, path, hash_dupe=hash_dupe)
        if res:
            doc2res(doc, path, info=info, hash_dupe=hash_dupe)
        if json:
            doc2json(doc, path, hash_dupe=hash_dupe)
        if pdb:
            doc2pdb(doc, path, hash_dupe=hash_dupe)
        if xsf:
            doc2xsf(doc, path)

    hull = subcmd in ['hull', 'voltage']
    if isinstance(cursor, pm.cursor.Cursor):
        cursor.rewind()
    md_path = "{directory}/{directory}.md".format(directory=directory)
    md_kwargs = {}
    md_kwargs.update(kwargs)
    md_kwargs.update({'markdown': True, 'latex': False, 'argstr': argstr, 'hull': hull})
    md_string = display_results(cursor, **md_kwargs)
    with open(md_path, 'w') as f:
        f.write(md_string)

    if latex:
        if isinstance(cursor, pm.cursor.Cursor):
            cursor.rewind()
        tex_path = "{directory}/{directory}.tex".format(directory=directory)
        print('Writing LaTeX file', tex_path + '...')
        tex_kwargs = {}
        tex_kwargs.update(kwargs)
        tex_kwargs.update({'latex': True, 'markdown': False, 'argstr': argstr, 'hull': hull})
        tex_string = display_results(cursor, **tex_kwargs)
        with open(tex_path, 'w') as f:
            f.write(tex_string)

    print('Done!')


@file_writer_function
def doc2param(doc, path, overwrite=False, hash_dupe=False, spin=False):
    """ Write basic .param file from single doc.

    Parameters:
        doc (dict): the document to write out to file
        path (str): the desired path to file

    Keyword arguments:
        spin (bool): enforce breaking of spin symmetry to magic number of 5.
        overwrite (bool): whether or not to overwrite colliding files.
        hash_dupe (bool): whether or not to create a unique filename for
            any colliding files, or just skip writing them.

    Returns:
        list, str: list of strings to write to file, with implicit newlines,
            and the required file extension.

    """
    from matador.utils.castep_params import CASTEP_PARAMS, CASTEP_VERSION
    param_set = set(CASTEP_PARAMS)
    param_dict = dict()
    for param in [param for param in param_set if param in doc]:
        param_dict[param] = doc[param]

    flines = []
    flines.append('# Param file generated by matador (Matthew Evans 2016)\n# CASTEP version {} parameter set'
                  .format(CASTEP_VERSION))

    if isinstance(spin, bool):
        if spin:
            spin = 5
        else:
            spin = None

    skip_keywords = ['nbands', 'nelectrons']
    for kw in skip_keywords:
        if kw in doc:
            print('Skipping keyword {} as it was probably not desired...'.format(kw))
            param_dict.pop(kw)

    if spin is not None and not doc.get('spin_polarized'):
        param_dict['spin_polarized'] = True

    if param_dict.get('spin_polarized'):
        total_spin = None
        if 'atomic_init_spins' in doc:
            total_spin = sum([val for val in doc['atomic_init_spins'] if val])
        elif spin is not None:
            total_spin = spin

        if total_spin is not None:
            flines.append("{0:30}: {1}".format('spin', total_spin))

    for param in param_dict:
        if param not in ['source', 'devel_code']:
            if param in ['basis_precision'] and 'cut_off_energy' in param_dict:
                continue
            flines.append("{0:30}: {1}".format(param, param_dict[param]))

    if 'encapsulated' in doc and doc['encapsulated']:
        try:
            from implicit_cnts import implicit_cnt_params
            cnt_params = implicit_cnt_params(doc['cnt_radius'])
            flines.append('%BLOCK DEVEL_CODE')
            flines.append('ADD_EXT_LOCPOT: \"gaussian_cylinder\"')
            flines.append('1D_STRESS_TENSOR: True')
            flines.append('gaussian_cylinder_pot:')
            flines.append('V0 = {V0}\nradius = {radius}\nbroadening = {fwhm}\naxial = \"0 0 1\"\ncentre = \"0.5 0.5 0\"\n'
                          .format(**cnt_params))
            flines.append(':endgaussian_cylinder_pot')
            flines.append('%ENDBLOCK DEVEL_CODE')
        except ImportError:
            print('Failed to import implicit_cnt_params from pyairss, so not writing '
                  'implicit cnt parameters from .res file to .param file.')

    if 'devel_code' in param_dict:
        flines.append('\n%BLOCK DEVEL_CODE')
        flines.append(param_dict['devel_code'].strip())
        flines.append('%ENDBLOCK DEVEL_CODE')

    ext = 'param'

    return flines, ext


@file_writer_function
def doc2cell(doc, path, overwrite=False, hash_dupe=False, spin=False):
    """ Write .cell file for single doc.

    Parameters:
        doc (dict): the document to write to file
        path (str): the desired path to the file

    Keyword Arguments:
        overwrite (bool): whether or not to overwrite colliding files.
        hash_dupe (bool): whether or not to create a unique filename for
            any colliding files, or just skip writing them.
        spin (bool): break spin symmetry with magic number 5 on first atom

    Returns:
        list, str: list of strings to write to file, with implicit newlines,
            and the required file extension.

    """
    if isinstance(spin, bool):
        if spin:
            spin = 5
        else:
            spin = None

    # if spin symmetry breaking is requested, update atomic init spins
    if 'positions_frac' in doc and len(doc['positions_frac']) > 0:
        if spin or 'atomic_init_spins' in doc:
            if not any(doc.get('atomic_init_spins', [])):
                doc['atomic_init_spins'] = len(doc['positions_frac']) * [None]
                doc['atomic_init_spins'][0] = spin

    if 'atomic_init_spins' in doc and len(doc['atomic_init_spins']) != len(doc['positions_frac']):
        print('Length mismatch between atoms and spins, will pad with zeros...')

    if 'site_occupancy' in doc:
        for occ in doc['site_occupancy']:
            if abs(occ - 1) > EPS:
                raise RuntimeError('Partial occupancies (VCA) not supported in cell files with matador')

    flines = []
    flines.append('# Cell file generated by matador (Matthew Evans 2016)\n')

    if 'lattice_cart' in doc:
        flines.append('%BLOCK LATTICE_CART')
        for vec in doc['lattice_cart']:
            flines.append('{d[0]} {d[1]} {d[2]}'.format(d=vec))
        flines.append('%ENDBLOCK LATTICE_CART')

    if 'atom_types' and ('positions_frac' in doc or 'positions_abs' in doc):

        if 'positions_frac' in doc:
            title = 'POSITIONS_FRAC'
            positions = doc['positions_frac']
        else:
            title = 'POSITIONS_ABS'
            positions = doc['positions_abs']

        flines.append('\n%BLOCK {}'.format(title))
        for ind, atom in enumerate(zip(doc['atom_types'], positions)):
            postfix = ''
            try:
                if 'atomic_init_spins' in doc and doc['atomic_init_spins'][ind]:
                    postfix = "SPIN={}".format(doc['atomic_init_spins'][ind])
            except IndexError:
                pass

            flines.append("{0:8s} {1[0]: 20.15f} {1[1]: 20.15f} {1[2]: 20.15f} {2:}".format(
                atom[0], atom[1], postfix))

        flines.append('%ENDBLOCK {}'.format(title))

    if 'external_pressure' in doc:
        flines.append('\n%BLOCK EXTERNAL_PRESSURE')
        if np.asarray(doc['external_pressure']).shape == (3, 3):
            flines.append('{d[0][0]} {d[0][1]} {d[0][2]}'.format(d=doc['external_pressure']))
            flines.append('{d[1][1]} {d[1][2]}'.format(d=doc['external_pressure']))
            flines.append('{d[2][2]}'.format(d=doc['external_pressure']))
        else:
            flines.append('{d[0][0]} {d[0][1]} {d[0][2]}'.format(d=doc['external_pressure']))
            flines.append('{d[1][0]} {d[1][1]}'.format(d=doc['external_pressure']))
            flines.append('{d[2][0]}'.format(d=doc['external_pressure']))

        flines.append('%ENDBLOCK EXTERNAL_PRESSURE')

    if 'ionic_constraints' in doc:
        flines.append('\n%BLOCK IONIC_CONSTRAINTS')
        for constraint in doc['ionic_constraints']:
            flines.append('{}'.format(constraint))
        flines.append('%ENDBLOCK IONIC_CONSTRAINTS')

    # specify SCF kpoints
    if 'kpoints_list' in doc:
        flines.append('\n%BLOCK KPOINTS_LIST')
        for kpoint in doc['kpoints_list']:
            flines.append('{p[0]:f} {p[1]:f} {p[2]:f} {p[3]:f}'.format(p=kpoint))
        flines.append('\n%ENDBLOCK KPOINTS_LIST')
    elif 'kpoints_mp_spacing' in doc:
        flines.append('\nKPOINTS_MP_SPACING: {}'.format(doc['kpoints_mp_spacing']))
    elif 'kpoints_mp_grid' in doc:
        flines.append('\nKPOINTS_MP_GRID: {d[0]} {d[1]} {d[2]}'.format(d=doc['kpoints_mp_grid']))

    if 'supercell_kpoints_mp_spacing' in doc:
        flines.append('\nSUPERCELL_KPOINTS_MP_SPACING: {}'.format(doc['supercell_kpoints_mp_spacing']))

    if 'kpoints_mp_offset' in doc:
        flines.append('\nKPOINTS_MP_OFFSET: {d[0]} {d[1]} {d[2]}'
                      .format(d=doc['kpoints_mp_offset']))

    if 'spectral_kpoints_mp_offset' in doc:
        flines.append('SPECTRAL_KPOINTS_MP_OFFSET: {d[0]} {d[1]} {d[2]}'
                      .format(d=doc['spectral_kpoints_mp_offset']))
    if 'phonon_kpoint_mp_offset' in doc:
        flines.append('PHONON_KPOINT_MP_OFFSET: {d[0]} {d[1]} {d[2]}'
                      .format(d=doc['phonon_kpoint_mp_offset']))

    # specify spectral kpoints
    if 'spectral_kpoints_mp_spacing' in doc:
        flines.append('\nSPECTRAL_KPOINTS_MP_SPACING: {}'
                      .format(doc['spectral_kpoints_mp_spacing']))
    elif 'spectral_kpoints_mp_grid' in doc:
        flines.append('\nSPECTRAL_KPOINTS_MP_GRID: {d[0]} {d[1]} {d[2]}'
                      .format(d=doc['spectral_kpoints_mp_grid']))
    elif 'spectral_kpoints_list' in doc:
        flines.append('\n%BLOCK SPECTRAL_KPOINTS_LIST')
        for point in doc['spectral_kpoints_list']:
            flines.append('{p[0]} {p[1]} {p[2]}'.format(p=point))
        flines.append('%ENDBLOCK SPECTRAL_KPOINTS_LIST')
    elif 'spectral_kpoints_path' in doc:
        flines.append('\n%BLOCK SPECTRAL_KPOINTS_PATH')
        for ind, point in enumerate(doc['spectral_kpoints_path']):
            flines.append('{p[0]} {p[1]} {p[2]}'.format(p=point))
            if 'spectral_kpoints_path_labels' in doc:
                flines[-1] += ' ! {}'.format(doc['spectral_kpoints_path_labels'][ind])
        flines.append('%ENDBLOCK SPECTRAL_KPOINTS_PATH')
    if 'spectral_kpoints_path_spacing' in doc:
        flines.append('\nSPECTRAL_KPOINTS_PATH_SPACING: {}'.format(doc['spectral_kpoints_path_spacing']))

    # specify phonon kpoints
    if 'phonon_kpoint_list' in doc:
        flines.append('\n%BLOCK PHONON_KPOINT_LIST')
        for point in doc['phonon_kpoint_list']:
            flines.append('{p[0]} {p[1]} {p[2]}'.format(p=point))
        flines.append('%ENDBLOCK PHONON_KPOINT_LIST')
    elif 'phonon_kpoint_mp_spacing' in doc:
        flines.append('\nPHONON_KPOINT_MP_SPACING: {}'
                      .format(doc['phonon_kpoint_mp_spacing']))
    elif 'phonon_kpoint_mp_grid' in doc:
        flines.append('\nPHONON_KPOINT_MP_GRID: {d[0]} {d[1]} {d[2]}'
                      .format(d=doc['phonon_kpoint_mp_grid']))

    # specify phonon fine kpoints
    if 'phonon_fine_kpoint_list' in doc:
        flines.append('\n%BLOCK PHONON_FINE_KPOINT_LIST')
        for point in doc['phonon_fine_kpoint_list']:
            flines.append('{p[0]} {p[1]} {p[2]}'.format(p=point))
        flines.append('%ENDBLOCK PHONON_FINE_KPOINT_LIST')
    elif 'phonon_fine_kpoint_mp_spacing' in doc:
        flines.append('\nPHONON_FINE_KPOINT_MP_SPACING: {}'
                      .format(doc['phonon_fine_kpoint_mp_spacing']))
    elif 'phonon_fine_kpoint_mp_grid' in doc:
        flines.append('\nPHONON_FINE_KPOINT_MP_GRID: {d[0]} {d[1]} {d[2]}'
                      .format(d=doc['phonon_fine_kpoint_mp_grid']))
    elif 'phonon_fine_kpoint_path' in doc:
        flines.append('\n%BLOCK PHONON_FINE_KPOINT_PATH')
        for ind, point in enumerate(doc['phonon_fine_kpoint_path']):
            flines.append('{p[0]} {p[1]} {p[2]}'.format(p=point))
            if 'phonon_fine_kpoint_path_labels' in doc:
                flines[-1] += ' ! {}'.format(doc['phonon_fine_kpoint_path_labels'][ind])
        flines.append('%ENDBLOCK PHONON_FINE_KPOINT_PATH')
    if 'phonon_fine_kpoint_path_spacing' in doc:
        flines.append('\nPHONON_FINE_KPOINT_PATH_SPACING: {}'.format(doc['phonon_fine_kpoint_path_spacing']))

    if 'phonon_supercell_matrix' in doc:
        flines.append('\n%BLOCK PHONON_SUPERCELL_MATRIX')
        for i in range(3):
            flines.append('{d[0]:3d} {d[1]:3d} {d[2]:3d}'.format(d=doc['phonon_supercell_matrix'][i]))
        flines.append('%ENDBLOCK PHONON_SUPERCELL_MATRIX')

    if 'cell_constraints' in doc:
        flines.append('\n%BLOCK CELL_CONSTRAINTS')
        flines.append('{d[0]} {d[1]} {d[2]}'.format(d=doc['cell_constraints'][0]))
        flines.append('{d[0]} {d[1]} {d[2]}'.format(d=doc['cell_constraints'][1]))
        flines.append('%ENDBLOCK CELL_CONSTRAINTS')

    if doc.get('fix_com'):
        flines.append('FIX_COM: TRUE')
    if doc.get('fix_all_cell'):
        flines.append('FIX_ALL_CELL: TRUE')
    if doc.get('fix_all_ions'):
        flines.append('FIX_ALL_IONS: TRUE')
    if doc.get('fix_vol'):
        flines.append('FIX_VOL: TRUE')

    if 'symmetry_generate' in doc:
        flines.append('SYMMETRY_GENERATE')
    if 'snap_to_symmetry' in doc:
        flines.append('SNAP_TO_SYMMETRY')
    if 'symmetry_tol' in doc:
        flines.append('SYMMETRY_TOL: {}'.format(doc['symmetry_tol']))
    if 'hubbard_u' in doc:
        flines.append('\n%BLOCK HUBBARD_U')
        flines.append('eV')
        for elem in doc['hubbard_u']:
            for orbital in doc['hubbard_u'][elem]:
                shift = str(doc['hubbard_u'][elem][orbital])
                flines.append('{} {} {}'.format(elem, orbital, shift))
        flines.append('%ENDBLOCK HUBBARD_U')
    if 'quantisation_axis' in doc:
        flines.append('QUANTISATION_AXIS: ')
        for integer in doc['quantisation_axis']:
            flines.append(str(integer) + ' ')
    if 'positions_noise' in doc:
        flines.append('POSITIONS_NOISE: {}'.format(doc['positions_noise']))
    if 'cell_noise' in doc:
        flines.append('CELL_NOISE : {}'.format(doc['cell_noise']))
    if 'species_pot' in doc:
        flines.append('\n%BLOCK SPECIES_POT')
        for elem in doc['species_pot']:
            if elem == 'library':
                flines.append(doc['species_pot']['library'])
            elif 'atom_types' not in doc or elem in doc.get('atom_types'):
                flines.append('{:4} {}'.format(elem, doc['species_pot'][elem]))
        flines.append('%ENDBLOCK SPECIES_POT')

    ext = 'cell'

    return flines, ext


def doc2pdb(doc, path, info=True, hash_dupe=True):
    """ Write a simple .pdb for single doc.

    Parameters:
        doc (dict): matador document containing structure.
        path (str): desired path of xsf file.

    Keyword arguments:
        info (bool): write info string to HEADER.
        hash_dupe (bool): hash duplicate file names or skip?

    """
    if path.endswith('.pdb'):
        path = path.replace('.pdb', '')
    try:
        if os.path.isfile(path+'.pdb'):
            if hash_dupe:
                print('File already exists, generating hash...')
                path += '-' + generate_hash()
            else:
                raise RuntimeError('Skipping duplicate structure...')
        with open(path+'.pdb', 'w') as f:
            try:
                header = 'HEADER    {} {}'.format(doc['text_id'][0], doc['text_id'][1])
            except Exception:
                header = 'HEADER    Generated with matador.'
            try:
                # write res file header if info
                title = 'TITLE     '
                title += path.split('/')[-1] + ' '
                if not doc.get('pressure'):
                    title += '0.00 '
                else:
                    title += str(doc['pressure']) + ' '
                title += str(doc['cell_volume']) + ' '
                title += str(doc['enthalpy']) + ' '
                title += '0 0 '             # spin
                title += str(doc['num_atoms']) + ' '
                try:
                    if 'x' in doc['space_group']:
                        title += '(P1) '
                    else:
                        title += '(' + str(doc['space_group']) + ')' + ' '
                except Exception:
                    title += '(P1) '
                title += 'n - 1'
            except Exception:
                if not info:
                    title = 'TITLE\t' + path.split('/')[-1]
                raise RuntimeError('Failed to get info for res file, turn info off.')
            author = 'AUTHOR    Generated with matador (Matthew Evans, 2016)'
            f.write(header + '\n')
            f.write(author + '\n')
            f.write(title + '\n')
            # use dummy SG for CRYST1, shouldn't matter
            cryst = ('CRYST1 {v[0][0]:9.3f} {v[0][1]:9.3f} {v[0][2]:9.3f} {v[1][0]:7.2f} {v[1][1]:7.2f} {v[1][2]:7.2f} P 1'
                     .format(v=doc['lattice_abc']))
            f.write(cryst + '\n')
            scale_n = cart2abcstar(doc['lattice_cart'])
            f.write('SCALE1    {v[0][0]:10.6f} {v[0][1]:10.6f} {v[0][2]:10.6f}      {:10.5f}\n'.format(0.0, v=scale_n))
            f.write('SCALE2    {v[1][0]:10.6f} {v[1][1]:10.6f} {v[1][2]:10.6f}      {:10.5f}\n'.format(0.0, v=scale_n))
            f.write('SCALE3    {v[2][0]:10.6f} {v[2][1]:10.6f} {v[2][2]:10.6f}      {:10.5f}\n'.format(0.0, v=scale_n))
            if 'positions_abs' not in doc:
                doc['positions_abs'] = frac2cart(doc['lattice_cart'], doc['positions_frac'])
            for ind, atom in enumerate(doc['atom_types']):
                try:
                    hetatm = 'HETATM '
                    # append 00 to atom type, a la cell2pdb...
                    hetatm += '{:4d} {:.4} NON A   1     '.format(ind+1, atom+'00')
                    hetatm += ('{v[0]:7.3f} {v[1]:7.3f} {v[2]:7.3f} {:5.2f} {:5.2f}          {:.2}'
                               .format(1.0, 0.0, atom, v=doc['positions_abs'][ind]))
                    f.write(hetatm + '\n')
                except Exception:
                    print_exc()
            ter = 'TER       {}       NON A   1'.format(len(doc['atom_types']))
            f.write(ter + '\n')
            f.write('END')
    except Exception:
        if hash_dupe:
            print_exc()
            print('Writing pdb file failed for ', doc['text_id'])
        else:
            print_exc()


def doc2json(doc, path, overwrite=False, hash_dupe=True):
    """ Return raw JSON document as stored in database.

    Parameters:
        doc (dict): matador document containing structure
        path (str): desired filename for res file

    Keyword arguments:
        hash_dupe (bool): hash duplicate file names, or skip?
        overwrite (bool): overwrite if filename exists.

    """
    try:
        import json
        if path.endswith('.json'):
            path = path.replace('.json', '')

        if os.path.isfile(path+'.json'):
            if overwrite:
                os.remove(path + '.json')
            elif hash_dupe:
                print('File name already exists, generating hash...')
                path += '-' + generate_hash()
            else:
                print('File name already exists! Skipping!')

        if '_id' in doc:
            doc['_id'] = str(doc['_id'])

        with open(path + '.json', 'w') as f:
            f.write(json.dumps(doc, skipkeys=True, indent=2))
    except Exception:
        print_exc()
        print('Writing {}.json failed!'.format(path))


def doc2pwscf(doc, path, template=None, spacing=None):
    """ Write the structural part of QE input file based
    on the provided matador doc. Will calculate the correct
    kpoint_mp_grid if spacing is provided.

    Parameters:
        doc (dict): matador document containing structure
        path (str): desired filename for res file

    Keyword Arguments:
        template (str): filename of template to prepend before structure
            (with prefix keyword to be replaced),
        spacing (float): kpoint_mp_spacing to use when calculating grid.

    """
    if path.endswith('.in'):
        path = path.replace('.in', '')

    if os.path.isfile(path + '.in'):
        print('File already exists, not overwriting...')
        return

    if 'lattice_cart' not in doc:
        doc['lattice_cart'] = abc2cart(doc['lattice_abc'])

    if 'kpoints_mp_spacing' in doc or spacing is not None:
        if 'kpoints_mp_spacing' in doc:
            spacing = doc['kpoints_mp_spacing']
        doc['kpoints_mp_grid'] = calc_mp_grid(doc['lattice_cart'], spacing)

    print(doc['kpoints_mp_grid'])

    file_string = ''
    file_string += 'CELL_PARAMETERS angstrom\n'
    for i in range(3):
        file_string += ' {d[0]: 10.10f} {d[1]: 10.10f} {d[2]: 10.10f}\n'.format(d=doc['lattice_cart'][i])

    file_string += '\n ATOMIC_POSITIONS crystal\n'
    for i in range(len(doc['atom_types'])):
        file_string += ('{:4} {d[0]: 10.10f} {d[1]: 10.10f} {d[2]: 10.10f}\n'
                        .format(doc['atom_types'][i], d=doc['positions_frac'][i]))
    file_string += '\nK_POINTS automatic\n'
    file_string += '{d[0]} {d[1]} {d[2]} 0 0 0'.format(d=doc['kpoints_mp_grid'])

    if template is not None:
        if os.path.isfile(template):
            with open(template, 'r') as f:
                template_string = f.readlines()

    with open(path + '.in', 'w') as f:
        for line in template_string:
            if 'prefix' in line:
                line = '  prefix = \'{}\',\n'.format(path)
            elif 'nat' in line:
                line = '  nat = {},\n'.format(len(doc['atom_types']))
            elif 'nat' in line:
                line = '  ntyp = {},\n'.format(len(set(doc['atom_types'])))
            f.write(line)
        f.write(file_string)


@file_writer_function
def doc2res(doc, path, overwrite=False, hash_dupe=False, info=True, spoof_titl=False, sort_atoms=True):
    """ Write .res file for single doc.

    Parameters:
        doc (dict): matador document containing structure
        path (str): desired filename for res file

    Keyword Arguments:
        info (bool): require info in res file header
        spoof_titl (bool): make up fake info for file header (for use with e.g. cryan)
        sorted (bool): if False, atoms are not sorted (this will not be a valid res file)
        overwrite (bool): whether or not to overwrite colliding files.
        hash_dupe (bool): whether or not to create a unique filename for
            any colliding files, or just skip writing them.

    """
    if spoof_titl:
        info = False
    if not info:
        space_group = 'P1' if 'space_group' not in doc else doc['space_group']
        if spoof_titl:
            titl = ('TITL {} -1 1 -1 0 0 {} ({}) n - 1').format(path.split('/')[-1], doc['num_atoms'], space_group)
        else:
            titl = ('TITL file generated by matador (Matthew Evans 2016)')
    else:
        try:
            titl = 'TITL '
            titl += (path.split('/')[-1] + ' ')
            if 'pressure' not in doc or isinstance(doc['pressure'], str):
                titl += '0.00 '
            else:
                titl += str(doc['pressure']) + ' '
            if 'cell_volume' not in doc:
                titl += '0.0 '
            else:
                titl += str(doc['cell_volume']) + ' '
            if 'enthalpy' in doc and not isinstance(doc['enthalpy'], str):
                titl += str(doc['enthalpy']) + ' '
            elif '0K_energy' in doc:
                titl += str(doc['0K_energy']) + ' '
            elif 'total_energy' in doc:
                titl += str(doc['total_energy']) + ' '
            else:
                raise KeyError('No energy field found.')
            titl += '0 0 '             # spin
            titl += str(doc['num_atoms']) + ' '
            if 'space_group' not in doc:
                titl += '(P1) '
            elif 'x' in doc['space_group']:
                titl += '(P1) '
            else:
                titl += '(' + str(doc['space_group']) + ')' + ' '
            titl += 'n - 1'
        except Exception:
            raise RuntimeError('Failed to get info for res file, turn info off.')
    if 'encapsulated' in doc and doc['encapsulated']:
        rem = ("REM NTPROPS {{\'chiralN\': {}, \'chiralM\': {}, \'r\': {}, "
               "\'offset\': [0.5, 0.5, 0.5], \'date\': \'xxx\', \'eformperfu\': 12345, \'z\': {}}}"
               .format(doc['cnt_chiral'][0], doc['cnt_chiral'][1], doc['cnt_radius'], doc['cnt_length']))
    flines = []
    if 'encapsulated' in doc and doc['encapsulated']:
        flines.append(rem)
    flines.append(titl)
    cell_str = 'CELL 1.0 '
    if ('lattice_abc' not in doc or len(doc['lattice_abc']) != 2
            or len(doc['lattice_abc'][0]) != 3 or len(doc['lattice_abc'][1]) != 3):
        try:
            doc['lattice_abc'] = cart2abc(doc['lattice_cart'])
        except Exception:
            raise RuntimeError('Failed to get lattice, something has gone wrong for {}'.format(path))
    for vec in doc['lattice_abc']:
        for coeff in vec:
            cell_str += ('{:.12f} '.format(coeff))
    flines.append(cell_str)
    flines.append('LATT -1')

    # enforce correct order by elements, sorting only the atom_types, not the positions inside them
    if len(doc['positions_frac']) != len(doc['atom_types']):
        raise RuntimeError('Atom/position array mismatch!')

    if 'site_occupancy' in doc:
        if len(doc['site_occupancy']) != len(doc['positions_frac']):
            raise RuntimeError('Occupancy/position array mismatch!')

    if 'site_occupancy' not in doc:
        occupancies = [1.0] * len(doc['positions_frac'])

    if sort_atoms:
        positions_frac, atom_types = zip(*[(pos, types) for (types, pos) in
                                           sorted(zip(doc['atom_types'], doc['positions_frac']),
                                                  key=lambda k: k[0])])
        if 'site_occupancy' in doc:
            occupancies, _atom_types = zip(*[(occ, types) for (types, occ) in
                                             sorted(zip(doc['atom_types'], doc['site_occupancy']),
                                                    key=lambda k: k[0])])

    else:
        positions_frac = doc['positions_frac']
        atom_types = doc['atom_types']

    if len(positions_frac) != len(doc['positions_frac']) or len(atom_types) != len(doc['atom_types']):
        raise RuntimeError('Site occupancy mismatch!')

    written_atoms = []
    sfac_str = 'SFAC \t'
    for elem in atom_types:
        if elem not in written_atoms:
            sfac_str += ' ' + str(elem)
            written_atoms.append(str(elem))
    flines.append(sfac_str)
    atom_labels = []
    i = 0
    j = 1
    while i < len(atom_types):
        num = atom_types.count(atom_types[i])
        atom_labels.extend(num*[j])
        i += num
        j += 1

    for atom in zip(atom_types, atom_labels, positions_frac, occupancies):
        flines.append("{0:8s}{1:3d}{2[0]: 20.15f} {2[1]: 20.15f} {2[2]: 20.15f}  {3: 20.15f}".format(
            atom[0], atom[1], atom[2], atom[3]))
    flines.append('END')
    # very important newline for compatibliy with cryan
    # flines.append('')

    return flines, 'res'


def doc2xsf(doc, path, write_energy=False, write_forces=False, overwrite=False):
    """ Write an .xsf file for a matador document, with positions in
    Cartesian coordinates. Optionally, write the energy in a comment
    at the top of the file for use with aenet.

    Parameters:
        doc (dict): matador document containing structure.
        path (str): desired path of xsf file.

    Keyword arguments:
        write_energy (bool): whether or not to write total energy in a comment
            as the first line of the file.
        write_forces (bool): whether or not to write the forces on each atom.
        overwrite (bool): overwrite if file exists.

    """
    if path.endswith('.xsf'):
        path = path.replace('.xsf', '')

    flines = []
    if write_energy:
        if 'total_energy' in doc:
            flines.append('# total energy = {:10.8f} eV\n'.format(doc['total_energy']))
        else:
            raise RuntimeError("Failed to write energy in xsf file: key 'total_energy' missing from input.")

    flines.append('CRYSTAL')
    flines.append('PRIMVEC')
    if 'lattice_cart' in doc:
        for i in range(3):
            flines.append('\t\t{lat[0]: 10.8f}\t{lat[1]: 10.8f}\t{lat[2]: 10.8f}'.format(lat=doc['lattice_cart'][i]))
    else:
        raise RuntimeError("Failed to write lattice in xsf file: key 'lattice_cart' missing from input.")
    flines.append('PRIMCOORD')
    flines.append('{} {}'.format(doc['num_atoms'], 1))
    if 'positions_abs' not in doc:
        doc['positions_abs'] = frac2cart(doc['lattice_cart'], doc['positions_frac'])
    for ind, (atom, position) in enumerate(zip(doc['atom_types'], doc['positions_abs'])):
        flines.append('{:2}\t{pos[0]: 16.8f}\t{pos[1]: 16.8f}\t{pos[2]: 16.8f}'.format(atom, pos=position))
        if write_forces:
            if 'forces' in doc:
                flines[-1] += ('\t{f[0]: 16.8f}\t{f[1]: 16.8f}\t{f[2]: 16.8f}'.format(f=doc['forces'][ind]))
            else:
                raise RuntimeError("Failed to write forces in xsf file: key 'forces' missing from input.")

    if os.path.isfile(path + '.xsf'):
        if overwrite:
            os.remove(path + '.xsf')
        else:
            print('File name already exists! Skipping!')
            raise RuntimeError('Duplicate file!')

    with open(path + '.xsf', 'w') as f:
        for line in flines:
            f.write(line + '\n')


@file_writer_function
def doc2arbitrary(doc, path, overwrite=False, hash_dupe=False):
    """ Write a Python dictionary into a standard CASTEP-style
    keyword: value file.

    Parameters:
        doc (dict): contains key-value pairs.
        path (str): filename to write to.

    Keyword arguments:
        hash_dupe (bool): hash duplicate file names, or skip?
        overwrite (bool): overwrite if filename exists.

    Returns:
        list: list of strs to write to file.

    """
    ext = None
    flines = []
    output_doc = {}
    # sanitise dict to include only unique-by-case keys
    for key in doc:
        if key.startswith('_'):
            continue
        if key.lower() == 'source':
            continue
        if key.lower() in output_doc:
            raise Warning('Key {} defined multiple times.'.format(key))
        output_doc[key.lower()] = doc[key]

    for key in output_doc:
        flines.append("{}: {}".format(key, output_doc[key]))
    return flines, ext



""" Test hull routines. """

import unittest
import copy
from os.path import realpath
from glob import glob

import numpy as np

from matador.hull import QueryConvexHull
from matador.query import DBQuery
from matador.scrapers.castep_scrapers import res2dict, castep2dict
from matador.utils.chem_utils import get_concentration

# grab abs path for accessing test data
REAL_PATH = "/".join(realpath(__file__).split("/")[:-1]) + "/"


class HullTest(unittest.TestCase):
    """ Test Convex hull functionality. """

    def test_hull_from_file(self):
        """ Loading hull structures from files. """
        res_list = glob(REAL_PATH + "data/hull-KPSn-KP/*.res")
        self.assertEqual(
            len(res_list),
            87,
            "Could not find test res files, please check installation...",
        )
        cursor = [res2dict(res)[0] for res in res_list]
        hull = QueryConvexHull(cursor=cursor, elements=["K", "Sn", "P"], no_plot=True)
        self.assertEqual(len(hull.hull_cursor), 16)

    def test_hull_from_custom_chempots(self):
        """ Loading hull structures from files. """
        res_list = glob(REAL_PATH + "data/hull-KPSn-KP/*.res")
        self.assertEqual(
            len(res_list),
            87,
            "Could not find test res files, please check installation...",
        )
        cursor = [res2dict(res)[0] for res in res_list]
        cursor = [doc for doc in cursor if len(doc["stoichiometry"]) != 1]
        hull = QueryConvexHull(
            cursor=cursor,
            elements=["K", "Sn", "P"],
            no_plot=True,
            chempots=[-791.456765, -928.045026 / 2.0, 878.326441 / 4.0],
        )
        self.assertEqual(len(hull.hull_cursor), 16)

    def test_hull_with_crystal_models(self):
        """ Loading hull structures from files. """
        cursor, failures = res2dict(
            REAL_PATH + "data/hull-KPSn-KP/*.res", as_model=True
        )
        self.assertEqual(
            len(cursor),
            87,
            "Could not find all test res files, please check installation...",
        )
        cursor = [doc for doc in cursor if doc.num_elements != 1]
        hull = QueryConvexHull(
            cursor=cursor,
            elements=["K", "Sn", "P"],
            no_plot=True,
            chempots=[-791.456765, -928.045026 / 2.0, 878.326441 / 4.0],
        )
        self.assertEqual(len(hull.hull_cursor), 16)

    def test_hull_from_file_with_extraneous_elements(self):
        """ Loading hull structures from files with too many elements. """
        res_list = glob(REAL_PATH + "data/hull-KPSn-KP/*.res")
        cursor = [res2dict(res)[0] for res in res_list]
        hull = QueryConvexHull(
            cursor=cursor, elements=["K", "Sn"], no_plot=True, debug=True
        )
        self.assertEqual(len(hull.hull_cursor), 5)

    def test_binary_hull_distances(self):
        """ Test computing binary hull distances. """
        res_list = glob(REAL_PATH + "data/hull-KP-KSnP_pub/*.res")
        self.assertEqual(
            len(res_list),
            295,
            "Could not find test res files, please check installation...",
        )
        cursor = [res2dict(res)[0] for res in res_list]
        hull = QueryConvexHull(cursor=cursor, elements=["K", "P"], no_plot=True)

        hull_dist_test = np.loadtxt(REAL_PATH + "data/test_KP_hull_dist.dat")
        np.testing.assert_array_almost_equal(
            np.sort(hull_dist_test), np.sort(hull.hull_dist), decimal=3
        )

    def test_ternary_hull_distances(self):
        """ Test computing ternary hull distances. """
        res_list = glob(REAL_PATH + "data/hull-KPSn-KP/*.res")
        self.assertEqual(
            len(res_list),
            87,
            "Could not find test res files, please check installation...",
        )
        cursor = [res2dict(res)[0] for res in res_list]
        hull = QueryConvexHull(cursor=cursor, elements=["K", "Sn", "P"], no_plot=True)
        self.assertEqual(len(hull.hull_cursor), 16)
        self.assertEqual(len(hull.cursor), 87)
        for ind, doc in enumerate(hull.cursor):
            hull.cursor[ind]["filename"] = doc["source"][0].split("/")[-1]

        structures = np.loadtxt(REAL_PATH + "data/test_KSnP.dat")
        hull_dist_test = np.loadtxt(REAL_PATH + "data/test_KSnP_hull_dist.dat")
        precomp_hull_dist = hull.get_hull_distances(structures, precompute=True)
        no_precomp_hull_dist = hull.get_hull_distances(structures, precompute=False)
        np.testing.assert_array_almost_equal(
            np.sort(hull_dist_test), np.sort(hull.hull_dist), decimal=3
        )
        np.testing.assert_array_almost_equal(
            np.sort(hull.hull_dist), np.sort(precomp_hull_dist), decimal=3
        )
        np.testing.assert_array_almost_equal(
            no_precomp_hull_dist, precomp_hull_dist, decimal=5
        )
        self.assertFalse(np.isnan(hull.hull_dist).any())

    def test_toy_ternary(self):
        cursor = [
            {
                "stoichiometry": [["K", 1.0]],
                "enthalpy_per_atom": 0,
                "cell_volume": 100,
                "num_atoms": 10,
            },
            {
                "stoichiometry": [["Sn", 1.0]],
                "enthalpy_per_atom": 0,
                "cell_volume": 100,
                "num_atoms": 10,
            },
            {
                "stoichiometry": [["P", 1.0]],
                "enthalpy_per_atom": 0,
                "cell_volume": 100,
                "num_atoms": 10,
            },
            {
                "stoichiometry": [["K", 1.0], ["Sn", 1.0], ["P", 1.0]],
                "enthalpy_per_atom": -1,
                "cell_volume": 100,
                "num_atoms": 30,
            },
            {
                "stoichiometry": [["K", 1.0], ["Sn", 1.0], ["P", 1.0]],
                "enthalpy_per_atom": -0.5,
                "cell_volume": 100,
                "num_atoms": 30,
            },
            {
                "stoichiometry": [["Sn", 1.0], ["P", 1.0]],
                "enthalpy_per_atom": 0.1,
                "cell_volume": 100,
                "num_atoms": 30,
            },
            {
                "stoichiometry": [["K", 4.0], ["Sn", 2.0], ["P", 2.0]],
                "enthalpy_per_atom": -0.74,
                "cell_volume": 100,
                "num_atoms": 30,
            },
            {
                "stoichiometry": [["K", 2.0], ["Sn", 2.0], ["P", 4.0]],
                "enthalpy_per_atom": -0.74,
                "cell_volume": 100,
                "num_atoms": 30,
            },
            {
                "stoichiometry": [["K", 2.0], ["Sn", 4.0], ["P", 2.0]],
                "enthalpy_per_atom": -0.74,
                "cell_volume": 100,
                "num_atoms": 30,
            },
            {
                "stoichiometry": [["Sn", 1.0], ["P", 1.0]],
                "enthalpy_per_atom": 0,
                "cell_volume": 100,
                "num_atoms": 30,
            },
        ]
        for ind, doc in enumerate(cursor):
            cursor[ind]["concentration"] = get_concentration(doc, ["K", "Sn", "P"])
            cursor[ind]["source"] = ["abcde"]
            cursor[ind]["num_fu"] = 10
            cursor[ind]["enthalpy"] = (
                cursor[ind]["enthalpy_per_atom"] * doc["num_atoms"]
            )

        hull = QueryConvexHull(cursor=cursor, elements=["K", "Sn", "P"], no_plot=True)

        hull_dists = [doc["hull_distance"] for doc in hull.cursor]
        np.testing.assert_array_almost_equal(
            hull_dists, [0, 0, 0, 0, 0.5, 0.1, 0.01, 0.01, 0.01, 0]
        )

    def test_pseudoternary_hull(self):
        cursor, s = res2dict(REAL_PATH + "data/hull-LLZO/*.res")
        print()
        print(80 * "-")
        self.assertEqual(
            len(cursor), 12, "Error with test res files, please check installation..."
        )
        hull = QueryConvexHull(
            cursor=cursor, elements=["La2O3", "ZrO2", "Li2O"], no_plot=True
        )
        self.assertEqual(len(hull.cursor), 7)
        self.assertEqual(len(hull.hull_cursor), 5)
        hull = QueryConvexHull(
            cursor=cursor,
            elements=["La2O3", "ZrO2", "Li2O"],
            hull_cutoff=0.01,
            no_plot=True,
        )
        self.assertEqual(len(hull.cursor), 7)
        self.assertEqual(len(hull.hull_cursor), 6)
        self.assertEqual(len(hull.convex_hull.vertices), 5)
        for doc in hull.hull_cursor:
            if "cubic-LLZO" in doc["source"][0]:
                self.assertAlmostEqual(
                    doc["formation_enthalpy_per_atom"], -0.05758265622
                )
                self.assertAlmostEqual(doc["hull_distance"], 0.008746875)
                break
        else:
            raise RuntimeError("Did not find cubic-LLZO in cursor")

    def test_pseudoternary_hull_NaVSO4(self):
        cursor, s = castep2dict(REAL_PATH + "data/hull-NaVSO4/*.castep")
        print(80 * "-")
        self.assertEqual(
            len(cursor), 9, "Error with test castep files, please check installation..."
        )
        hull = QueryConvexHull(
            cursor=cursor,
            elements=["Na", "V", "SO4"],
            chempots=[-999, -999, 100],
            no_plot=True,
        )
        self.assertEqual(len(hull.cursor), 8)
        self.assertEqual(len(hull.hull_cursor), 6)

    def test_pseudoternary_hull_with_custom_chempots(self):
        cursor, s = res2dict(REAL_PATH + "data/hull-LLZO/*.res")
        print()
        print(80 * "-")
        self.assertEqual(
            len(cursor), 12, "Error with test res files, please check installation..."
        )
        hull = QueryConvexHull(
            cursor=cursor,
            elements=["La2O3", "ZrO2", "Li2O"],
            chempots=[-3275.040 / 5, -2178.987 / 3, -848.148 / 3],
            no_plot=True,
        )
        self.assertEqual(len(hull.cursor), 10)
        self.assertEqual(len(hull.hull_cursor), 5)

        # cursor, s = res2dict(REAL_PATH + 'data/hull-LLZO/*.res')
        hull = QueryConvexHull(
            cursor=cursor,
            elements=["La2O3", "ZrO2", "Li2O"],
            hull_cutoff=0.20,
            chempots=[26200.3194 / 40, -8715.94784 / 12, -3392.59361 / 12],
            no_plot=True,
        )
        self.assertEqual(len(hull.cursor), 10)
        self.assertEqual(len(hull.hull_cursor), 10)

        for doc in hull.hull_cursor:
            if "cubic-LLZO" in doc["source"][0]:
                self.assertAlmostEqual(
                    doc["formation_enthalpy_per_atom"], -0.05758265622
                )
                self.assertAlmostEqual(doc["hull_distance"], 0.008746875)
                break
        else:
            raise RuntimeError("Did not find cubic-LLZO in cursor")

    def test_pseudoternary_hull_with_custom_chempots_below_hull(self):
        cursor, s = res2dict(REAL_PATH + "data/hull-LLZO/*.res")
        hull = QueryConvexHull(
            cursor=cursor,
            elements=["La2O3", "ZrO2", "Li2O"],
            hull_cutoff=0.02,
            chempots=[26200.3194 / 40, -8715.94784 / 12, -3392.59361 / 12],
            no_plot=True,
        )
        self.assertEqual(len(hull.cursor), 10)
        self.assertEqual(len(hull.hull_cursor), 9)
        for doc in cursor:
            if "La2O3" in doc["source"][0]:
                new_doc = copy.deepcopy(doc)
                new_doc["enthalpy_per_atom"] -= 1
                new_doc["source"][0] = "IMAGINARY.res"
        hull_dists = hull.get_hull_distances([[1, 0, -1]])
        self.assertAlmostEqual(hull_dists[0], -1)

    def test_hull_pseudobinary_hull(self):
        cursor, s = res2dict(REAL_PATH + "data/query-SrTiO-oqmd_1.1/*.res")
        print()
        print(80 * "-")
        self.assertEqual(
            len(cursor), 202, "Error with test res files, please check installation..."
        )
        hull = QueryConvexHull(cursor=cursor, species=["Sr", "TiO3"], no_plot=True)

        self.assertEqual(len(hull.cursor), 37)
        self.assertEqual(len(hull.hull_cursor), 3)

        hull = QueryConvexHull(
            cursor=cursor, species=["Sr", "TiO3"], hull_cutoff=0.05, no_plot=True
        )
        self.assertEqual(len(hull.cursor), 37)
        self.assertEqual(len(hull.hull_cursor), 10)

    def test_pseudoternary_from_fake_query(self):
        cursor, s = res2dict(REAL_PATH + "data/hull-LLZO/*.res")
        print()
        print(80 * "-")
        self.assertEqual(
            len(cursor), 12, "Error with test res files, please check installation..."
        )
        hull = QueryConvexHull(
            cursor=cursor, elements=["La2O3", "ZrO2", "Li2O"], no_plot=True
        )
        self.assertEqual(len(hull.cursor), 7)

        fake_query = DBQuery.__new__(DBQuery)
        fake_query.cursor = hull.cursor
        for ind, doc in enumerate(fake_query.cursor):
            fake_query.cursor[ind]["_id"] = None
            fake_query.cursor[ind]["text_id"] = [doc["source"][0], "."]
        fake_query._non_elemental = True
        fake_query._create_hull = True
        fake_query.args = dict()
        fake_query.args["intersection"] = True
        fake_query.args["subcmd"] = "hull"
        fake_query.args["composition"] = ["La2O3:ZrO2:Li2O"]
        hull = QueryConvexHull(
            query=fake_query,
            hull_cutoff=0.01,
            chempots=[26200.3194 / 40, -8715.94784 / 12, -3392.59361 / 12],
            no_plot=True,
        )
        self.assertEqual(len(hull.cursor), 10)
        self.assertEqual(len(hull.hull_cursor), 9)

    def test_variations_of_element_parsing(self):
        cursor, s = res2dict(REAL_PATH + "data/hull-KP-KSnP_pub/*.res")
        self.assertEqual(
            len(cursor), 295, "Error with test res files, please check installation..."
        )
        msg = "failed to parse elements=['KP']"
        hull = QueryConvexHull(cursor=cursor, elements=["KP"], no_plot=True)
        self.assertEqual(len(hull.cursor), 295, msg=msg)
        self.assertEqual(len(hull.hull_cursor), 7, msg=msg)

        msg = "failed to parse species=['KP']"
        hull = QueryConvexHull(cursor=cursor, species=["KP"], no_plot=True)
        self.assertEqual(len(hull.cursor), 295, msg=msg)
        self.assertEqual(len(hull.hull_cursor), 7, msg=msg)

        msg = "failed to parse elements=['K', 'P']"
        hull = QueryConvexHull(cursor=cursor, elements=["K", "P"], no_plot=True)
        self.assertEqual(len(hull.cursor), 295, msg=msg)
        self.assertEqual(len(hull.hull_cursor), 7, msg=msg)

        msg = "failed to parse species=['K', 'P']"
        hull = QueryConvexHull(cursor=cursor, species=["K", "P"], no_plot=True)
        self.assertEqual(len(hull.cursor), 295, msg=msg)
        self.assertEqual(len(hull.hull_cursor), 7, msg=msg)

        msg = "failed to parse species='KP'"
        hull = QueryConvexHull(cursor=cursor, species="KP", no_plot=True)
        self.assertEqual(len(hull.cursor), 295, msg=msg)
        self.assertEqual(len(hull.hull_cursor), 7, msg=msg)

        msg = "failed to parse elements='KP'"
        hull = QueryConvexHull(cursor=cursor, elements="KP", no_plot=True)
        self.assertEqual(len(hull.cursor), 295, msg=msg)
        self.assertEqual(len(hull.hull_cursor), 7, msg=msg)

        msg = "failed to parse elements='K:P'"
        hull = QueryConvexHull(cursor=cursor, elements="K:P", no_plot=True)
        self.assertEqual(len(hull.cursor), 295, msg=msg)
        self.assertEqual(len(hull.hull_cursor), 7, msg=msg)

        msg = "failed to parse species='K:P'"
        hull = QueryConvexHull(cursor=cursor, species="K:P", no_plot=True)
        self.assertEqual(len(hull.cursor), 295, msg=msg)
        self.assertEqual(len(hull.hull_cursor), 7, msg=msg)

    def test_binary_from_fake_query(self):
        cursor, s = res2dict(REAL_PATH + "data/hull-KP-KSnP_pub/*.res")
        print()
        print(80 * "-")
        self.assertEqual(
            len(cursor), 295, "Error with test res files, please check installation..."
        )
        hull = QueryConvexHull(cursor=cursor, elements=["KP"], no_plot=True)
        self.assertEqual(len(hull.cursor), 295)
        self.assertEqual(len(hull.hull_cursor), 7)

        fake_query = DBQuery.__new__(DBQuery)
        fake_query.cursor = hull.cursor
        for ind, doc in enumerate(fake_query.cursor):
            fake_query.cursor[ind]["_id"] = None
            fake_query.cursor[ind]["text_id"] = [doc["source"][0], "."]
            del fake_query.cursor[ind]["hull_distance"]
            del fake_query.cursor[ind]["concentration"]

        fake_query._non_elemental = False
        fake_query._create_hull = True
        fake_query.args = dict()
        fake_query.args["intersection"] = False
        fake_query.args["subcmd"] = "hull"
        fake_query.args["composition"] = ["KP"]
        hull = QueryConvexHull(
            query=fake_query, chempots=[-791.456765, -219.58161025], no_plot=True
        )
        # now need to include custom chempots in counts
        self.assertEqual(len(hull.cursor), 297)
        self.assertEqual(len(hull.hull_cursor), 9)

    def test_filter_cursor(self):
        cursor, s = res2dict(REAL_PATH + "data/hull-LLZO/*.res")
        species = ["Li2O", "La2O3", "ZrO2"]
        new_cursor = QueryConvexHull.filter_cursor_by_chempots(species, cursor)
        self.assertEqual(len(cursor), 12)
        self.assertEqual(len(new_cursor), 7)
        cursor = [{"stoichiometry": [["Li", 7], ["O", 12], ["Zr", 2], ["La", 3]]}]
        new_cursor = QueryConvexHull.filter_cursor_by_chempots(species, cursor)
        self.assertEqual(len(new_cursor), 1)
        self.assertAlmostEqual(
            new_cursor[0]["concentration"][0], 0.5, msg="Concentrations do not match"
        )
        self.assertAlmostEqual(
            new_cursor[0]["concentration"][1],
            1.5 / 7.0,
            msg="Concentrations do not match",
        )

    def test_pseudoternary_hull_failure(self):
        cursor, s = res2dict(REAL_PATH + "data/hull-LLZO/*.res")
        print()
        print(80 * "-")
        self.assertEqual(
            len(cursor), 12, "Error with test res files, please check installation..."
        )
        with self.assertRaises(RuntimeError):
            QueryConvexHull(cursor=cursor, no_plot=True)


class EnsembleHullTest(unittest.TestCase):
    """ Test of Ensemble Hulls for BEEF/temperature. """

    def test_beef_hull(self):
        from matador.hull import EnsembleHull
        from matador.scrapers import castep2dict

        cursor, s = castep2dict(REAL_PATH + "data/beef_files/*.castep", db=False)

        beef_hull = EnsembleHull(
            cursor, "_beef", energy_key="total_energy_per_atom", parameter_key="thetas"
        )

        self.assertEqual(len(beef_hull.phase_diagrams), 5000)
        self.assertEqual(len(beef_hull.cursor[0]["_beef"]["hull_distance"]), 5000)
        self.assertEqual(len(beef_hull.cursor[1]["_beef"]["hull_distance"]), 5000)



# coding: utf-8
# Distributed under the terms of the MIT License.

""" Run calculations in a folder with BatchRun
such that there are no clashes.
"""
import os
import argparse
from matador import __version__, script_epilog
from matador.utils.print_utils import print_notify
from matador.utils.errors import InputError
from matador.compute import BatchRun


def main():
    """ Parse args and run any remaining jobs. """
    parser = argparse.ArgumentParser(
        prog='run3',
        description='Run multiple calculations from a series of .res \
                     files and single cell and param files, typically CASTEP geometry \
                     optimisations. The geometry optimization will \
                     be split into "--rough" (default: 4) chunks of "--rough_iter" (default: 2) \
                     iterations, followed by chunks of "--fine_iter" (default: 20) \
                     iterations, until geom_max_iter is reached in the param file. \
                     Successful runs will be moved to "completed", crashes/failures will go to \
                     "bad_castep" and initial res files will go into "input". Running jobs will \
                     be listed in jobs.txt and those that completed cleanly will be listed \
                     in finished_cleanly.txt.',
        epilog=script_epilog
    )
    parser.add_argument('--version', action='version', version='matador version ' + __version__ + '.')
    parser.add_argument('seed', type=str, nargs='+',
                        help='cell and param seed to use as template for CASTEP calculations OR list of files\
                              to apply run executable on')
    parser.add_argument('-nc', '--ncores', type=int,
                        help='number of cores per node per job [DEFAULT=cpu_count/nprocesses]')
    parser.add_argument('-np', '--nprocesses', type=int, default=1,
                        help='number of concurrent calculations, i.e. number \
                              of concurrent mpiruns [DEFAULT=1]')
    parser.add_argument('-nn', '--nnodes', type=int, default=1,
                        help='number of nodes per job, i.e. number of nodes \
                              using -nc cores [DEFAULT=1].')

    parser.add_argument('-t', '--max_walltime', type=int,
                        help='maximum walltime in seconds (job will quit early to clean up if specified)')
    parser.add_argument('-exec', '--executable', type=str,
                        help='specify path to or name of executable (DEFAULT: castep)')
    parser.add_argument('--no_reopt', action='store_true', default=False,
                        help='do not run geometry optimisation again after first success')
    parser.add_argument('--redirect', type=str,
                        help='filename to redirect output to, can use $seed macro')
    parser.add_argument('--mode', type=str, default='castep',
                        help='either castep or generic')
    parser.add_argument('--noise', action='store_true',
                        help=('add 0.1 A of random noise to positions on every cell, '
                              'useful for converging forces (DEFAULT: off)'))
    parser.add_argument('--squeeze', action='store_true',
                        help='add external pressure to the rough steps of geom opts')
    parser.add_argument('--ignore_jobs_file', action='store_true',
                        help='whether to use the jobs.txt file to avoid clashes')
    parser.add_argument('-d', '--debug', action='store_true', default=False,
                        help='debug output')
    parser.add_argument('-cust', '--custom_params', action='store_true', default=False,
                        help='use custom param file per structure')
    parser.add_argument('-v', '--verbosity', type=int, default=2,
                        help='integer to set level of verbosity')
    parser.add_argument('--archer', action='store_true', default=False,
                        help='use aprun over mpirun')
    parser.add_argument('--slurm', action='store_true', default=False,
                        help='use srun over mpirun')
    parser.add_argument('--intel', action='store_true', default=False,
                        help='use Intel\'s mpirun')
    parser.add_argument('--conv_cutoff', action='store_true', default=False,
                        help='run all res files at cutoff defined in cutoff.conv file')
    parser.add_argument('--conv_kpt', action='store_true', default=False,
                        help='run all res files at kpoint spacings defined in kpt.conv file')
    parser.add_argument('--memcheck', action='store_true', default=False,
                        help='enable memcheck via castep dryrun')
    parser.add_argument('--scratch_prefix', type=str,
                        help='specify absolute path prefix for compute dir e.g. '
                             '--scratch_prefix /scratch/user/ will set the compute directory to /scratch/user/$hostname. '
                             'default value is taken from .matadorrc.')
    parser.add_argument('--maxmem', type=int,
                        help='override max memory for memcheck')
    parser.add_argument('--killcheck', action='store_true', default=True,
                        help='check for $seed.kill file and quit job if present')
    parser.add_argument('--kpts_1D', action='store_true', default=False,
                        help='recalculate a 1D kpoint mesh of spacing specified in template cell')
    parser.add_argument('--spin', type=int, nargs='?', const=5, default=None,
                        help=('if not specified in .cell file, break spin symmetry on first atom using the spin specified by '
                              'the user [DEFAULT: 5]'))
    parser.add_argument('--rough', type=int, default=4,
                        help='choose how many <rough_iter> geometry optimizations \
                              to perform, decrease if lattice is nearly correct. [DEFAULT: 4].')
    parser.add_argument('--rough_iter', type=int, default=2,
                        help='choose how many relaxation steps per rough geometry optimization\
                              to perform, [DEFAULT: 2].')
    parser.add_argument('--fine_iter', type=int, default=20,
                        help='choose how many relaxation steps per fine geometry optimization\
                              to perform, [DEFAULT: 20].')
    parser.add_argument('-l', '--limit', type=int, default=None,
                        help='limit to n structures per run')
    parser.add_argument('--profile', action='store_true',
                        help='profile code with cProfile')
    args = parser.parse_args()

    seed = vars(args)['seed']

    kwargs = vars(args)
    del kwargs['seed']

    from matador.config import load_custom_settings
    settings = load_custom_settings(debug=kwargs.get('debug')).get('run3')
    if settings is not None:
        if kwargs['scratch_prefix'] is None:
            kwargs['scratch_prefix'] = settings.get('scratch_prefix')
            if kwargs['scratch_prefix'] == '.':
                kwargs['scratch_prefix'] = None

        if kwargs['executable'] is None:
            kwargs['executable'] = settings.get('castep_executable', 'castep')

        kwargs['run3_settings'] = settings

    if sum([vars(args)['slurm'], vars(args)['archer'], vars(args)['intel']]) > 1:
        exit('Incompatible MPI arguments specified, please use at most one of --archer/--intel/--slurm.')

    if vars(args).get('profile'):
        import cProfile
        import pstats
        from sys import version_info
        hostname = os.uname()[1]
        pr = cProfile.Profile()
        pr.enable()

    try:
        runner = BatchRun(seed, **kwargs)
        runner.spawn()
    except InputError as exc:
        print_notify(exc)
    except RuntimeError:
        print_notify('Some jobs failed, exiting...')
    except Exception as exc:
        raise exc

    if vars(args).get('profile'):
        pr.disable()
        fname = 'run3-{}-{}-{}.{}.{}'.format(__version__, hostname, version_info.major,
                                             version_info.minor, version_info.micro)
        pr.dump_stats(fname + '.prof')
        with open(fname + '.pstats', 'w') as s:
            sortby = 'cumulative'
            ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
            ps.print_stats()


if __name__ == '__main__':
    main()



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This file implements atomic swaps through the `AtomicSwapper` class. """


import re
from copy import deepcopy
from matador.utils.print_utils import print_success, print_warning
from matador.utils.chem_utils import get_periodic_table, get_stoich


class AtomicSwapper:
    """ This class handles the creation of input files from database
    queries that have swapped atoms.

    """
    def __init__(
            self, cursor, swap=None, uniq=False, top=None, maintain_num_species=True, debug=False, **kwargs):
        """ Initialise class with query cursor and arguments.

        Parameters:
            cursor (list): cursor of documents to swap.

        Keyword arguments:
            swap (str): specification of swaps to perform, e.g.
                "LiP:KSn" will swap all Li->P and all K->Sn in the
                cursor.
            uniq (bool/float): filter documents by similarity with
                the default sim_tol (True) or the value provided here.
            top (int): only swap from the first `top` structures in
                the cursor.
            maintain_num_species (bool): only perform swaps that maintain
                the number of species in the structure
            debug (bool): enable debug output
            kwargs (dict): dictionary of extra arguments that should be ignored.

        """
        # define some swap macros
        self.periodic_table = get_periodic_table()
        self.maintain_num_species = maintain_num_species
        self.swap_dict_list = None
        self.swap_args = swap
        del self.periodic_table['X']
        self.template_structure = None
        self.cursor = list(cursor)
        if top is not None:
            self.cursor = self.cursor[:top]

        if len(self.cursor) == 0:
            return

        self.swap_counter = 0
        self.parse_swaps(self.swap_args)
        swap_cursor = []
        for doc in self.cursor:
            docs, counter = self.atomic_swaps(doc)
            self.swap_counter += counter
            if counter > 0:
                swap_cursor.extend(docs)
        self.cursor = swap_cursor
        if self.swap_counter > 0:
            print_success('Performed {} swaps.'.format(self.swap_counter))
        else:
            print_warning('No swaps performed.')

        if uniq:
            from matador.utils.cursor_utils import filter_unique_structures
            print('Filtering for unique structures...')
            filtered_cursor = filter_unique_structures(self.cursor, debug=debug, sim_tol=uniq)
            print('Filtered {} down to {}'.format(len(self.cursor), len(filtered_cursor)))
            self.cursor = filtered_cursor

    def parse_swaps(self, swap_args=None):
        """ Parse command line options into valid atomic species swaps.

        e.g. --swap LiP:NaAs

            ==> [[['Li'], ['P']], [['Na'], ['P']].

        Handles multiple many-to-many swaps, macros for groups of the
        periodic table, and wildcards.

        Keyword arguments:
            swap_args (str): overrides command-line swap args.


        """

        self.swap_pairs = []

        if swap_args is None:
            swap_args = self.swap_args

        if swap_args is None:
            raise RuntimeError('No swap arguments passed.')

        if isinstance(swap_args, str):
            swap_args = [swap_args.strip()]

        if len(swap_args) > 1:
            raise RuntimeError('Detected whitespace in your input clear it and try again.')

        swap_list = swap_args[0].split(':')
        for swap in swap_list:
            if len(swap) <= 1:
                raise RuntimeError('Not enough arguments for swap!')
            # check is both options are groups
            if '][' in swap:
                tmp_list = [x for x in swap.split('][') if x != '']
            # check if only first option is group
            elif swap[0] == '[':
                tmp_list = [x for x in swap.split(']') if x != '']
            # check if only last option is group
            elif swap[-1] == ']':
                tmp_list = [x for x in swap.split('[') if x != '']
            # check if no groups
            else:
                tmp_list = [x for x in re.split(r'([A-Z][a-z]*)', swap) if x != '']
            for ind, tmp in enumerate(tmp_list):
                tmp_list[ind] = self._atoms_to_list(tmp)
            if len(tmp_list) != 2:
                raise RuntimeError('Unable to parse swap! {} should contain only two entries'.format(tmp_list))
            self.swap_pairs.append(tmp_list)
            self.construct_swap_options()

    def _atoms_to_list(self, atom_string):
        """ For a given set of atoms in a string, parse any macros and
        return a list of options.

        e.g. '[V' -> [<all group V atoms>],
        and 'V' -> ['V'].

        Parameters:
            atom_string (str): formula string with macros.

        """
        if '[' in atom_string or ']' in atom_string:
            group = atom_string.replace('[', '')
            group = group.replace(']', '')
            if group in self.periodic_table:
                atom_list = self.periodic_table[group]
            else:
                atom_list = group.split(',')
        else:
            return [atom_string]
        return [x.strip() for x in atom_list]

    def construct_swap_options(self):
        """ Iterate over possible combinations of multiple many-to-many
        swaps and create a dict for each swap.

        """
        self.swap_dict_list = []
        from itertools import product
        for branch in product(*([pair[1] for pair in self.swap_pairs])):
            self.swap_dict_list.append(dict())
            for ind, pair in enumerate(self.swap_pairs):
                for swap_from in pair[0]:
                    if swap_from != branch[ind]:
                        self.swap_dict_list[-1][swap_from] = branch[ind]

    def atomic_swaps(self, source_doc):
        """ Swap atomic species according to parsed options.

        Parameters:
            source_doc (dict): matador doc to swap from.

        """
        new_doc = deepcopy(source_doc)
        swapped_docs = []
        unswapped_num_species = len(set(source_doc['atom_types']))
        for swap in self.swap_dict_list:
            if any(key in source_doc['atom_types'] for key in swap):
                new_doc['atom_types'] = [swap.get(atom, atom) for atom in source_doc['atom_types']]
                new_doc['_swapped_stoichiometry'] = get_stoich(source_doc['atom_types'])
                new_doc['stoichiometry'] = get_stoich(new_doc['atom_types'])
                new_doc['elems'] = set(new_doc['atom_types'])
                new_doc['num_species'] = len(new_doc['elems'])
                if not self.maintain_num_species or new_doc['num_species'] == unswapped_num_species:
                    swapped_doc = deepcopy(new_doc)
                    swapped_docs.append(swapped_doc)

        return swapped_docs, len(swapped_docs)



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This submodule defines classes for computing, combining
and convolving pair distribution functions.

"""


from itertools import combinations_with_replacement
import itertools
import copy
from math import ceil, copysign
import time

import numpy as np
import numba

from matador.utils.cell_utils import frac2cart, cart2volume
from matador.utils.cell_utils import standardize_doc_cell
from matador.utils.chem_utils import get_stoich
from matador.fingerprints.fingerprint import Fingerprint, FingerprintFactory


class PDF(Fingerprint):
    """ This class implements the calculation and comparison of pair
    distribution functions.

    Attributes:
        r_space (ndarray) : 1-D array containing real space grid
        gr (ndarray): 1-D array containing total PDF
        dr (float): real-space grid spacing in Å
        rmax (float): extent of real-space grid in Å
        label (str): structure label
        elem_gr (dict): dict with pairs of element symbol keys,
            containing 1-D arrays of projected PDFs (if calculated)
        number_density (float): number density for renormalisation and
            comparison with other PDFs
        kwargs (dict): arguments used to create PDF

    """

    def __init__(self, doc, lazy=False, **kwargs):
        """ Initialise parameters and run PDF (unless lazy=True).

        Parameters:

            doc (dict) : matador document to calculate PDF of

        Keyword Arguments:

            dr (float) : bin width for PDF (Angstrom) (DEFAULT: 0.01)
            gaussian_width (float) : width of Gaussian smearing (Angstrom) (DEFAULT: 0.01)
            num_images (int/str) : number of unit cell images include in PDF calculation (DEFAULT: 'auto')
            max_num_images (int) : cutoff number of unit cells before crashing (DEFAULT: 50)
            rmax (float) : maximum distance cutoff for PDF (Angstrom) (DEFAULT: 15)
            projected (bool) : optionally calculate the element-projected PDF
            standardize (bool) : standardize cell before calculating PDF
            lazy (bool) : if True, calculator is not called when initializing PDF object
            timing (bool) : if True, print the total time taken to calculate the PDF

        """

        prop_defaults = {'dr': 0.01, 'gaussian_width': 0.1, 'rmax': 15, 'num_images': 'auto',
                         'style': 'smear', 'debug': False, 'timing': False, 'low_mem': False, 'projected': True,
                         'max_num_images': 50, 'standardize': True}

        # read and store kwargs
        self.kwargs = prop_defaults
        self.kwargs.update({key: kwargs[key]
                            for key in kwargs
                            if kwargs[key] is not None})

        # useful data for labelling
        self.spg = None
        structure = copy.deepcopy(doc)
        if self.kwargs.get('standardize'):
            structure = standardize_doc_cell(structure)
            self.spg = structure['space_group']
        self.stoichiometry = structure.get('stoichiometry', get_stoich(structure['atom_types']))

        # private variables
        self._num_images = self.kwargs.get('num_images')
        self._lattice = np.asarray(structure['lattice_cart'])
        self._poscart = np.asarray(frac2cart(structure['lattice_cart'], structure['positions_frac'])).reshape(-1, 3)
        self._types = structure['atom_types']
        self._num_atoms = len(self._poscart)
        self._volume = cart2volume(self._lattice)
        self._image_vec = None

        # public variables
        self.rmax = self.kwargs.get('rmax')
        self.number_density = self._num_atoms / self._volume
        self.dr = self.kwargs.get('dr')
        self.r_space = None
        self.gr = None
        self.elem_gr = None

        self.label = None
        if self.kwargs.get('label'):
            self.label = self.kwargs["label"]
        elif 'text_id' in structure:
            self.label = ' '.join(structure['text_id'])

        if not lazy:
            if self.kwargs.get('timing'):
                start = time.time()
            self.calc_pdf()
            if self.kwargs.get('timing'):
                end = time.time()
                print('PDF calculated in {:.3f} s'.format(end - start))

    def calc_pdf(self):
        """ Wrapper to calculate PDF with current settings. """
        if self._image_vec is None:

            if self.kwargs.get('debug'):
                start = time.time()

            self._set_image_trans_vectors()

            if self.kwargs.get('debug'):
                end = time.time()
                print('Image vectors length = {}'.format(len(self._image_vec)))
                print('Set image trans vectors in {} s'.format(end - start))

        if self.kwargs.get('projected'):
            if self.kwargs.get('debug'):
                start = time.time()

            if self.elem_gr is None:
                self._calc_projected_pdf()

            if self.kwargs.get('debug'):
                end = time.time()
                print('Calculated projected PDF {} s'.format(end - start))
        if self.gr is None:
            self._calc_unprojected_pdf()

    def calculate(self):
        """ Alias for `self.calc_pdf`. """
        self.calc_pdf()

    def _calc_distances(self, poscart, poscart_b=None):
        """ Calculate PBC distances with cdist.

        Parameters:
            poscart (numpy.ndarray): array of absolute atomic coordinates.

        Keyword Arguments:
            poscart_b (numpy.ndarray): absolute positions of a second type of atoms,
                where only A-B distances will be calculated.

        Returns:
            numpy.ndarray: pair d_ij matrix with values > rmax < 1e-12 removed.

        """
        from matador.utils.cell_utils import calc_pairwise_distances_pbc
        return calc_pairwise_distances_pbc(poscart,
                                           self._image_vec,
                                           self._lattice,
                                           self.rmax,
                                           filter_zero=True,
                                           compress=True,
                                           poscart_b=poscart_b,
                                           debug=self.kwargs.get('debug'))

    def _calc_unprojected_pdf(self):
        """ Wrapper function to calculate distances and output
        a broadened and normalised PDF. Sets self.gr and self.r_space
        to G(r) and r respectively.

        """
        if self.elem_gr is not None:
            self._calc_unprojected_pdf_from_projected()
        else:
            distances = self._calc_distances(self._poscart)
            self.r_space = np.arange(0, self.rmax + self.dr, self.dr)
            self.gr = self._get_broadened_normalised_pdf(distances,
                                                         style=self.kwargs.get('style'),
                                                         gaussian_width=self.kwargs.get('gaussian_width'))

    def _calc_projected_pdf(self):
        """ Calculate broadened and normalised element-projected PDF of a matador document.
        Sets self.elem_gr of e.g. Li2Zn3 to

            {
                ('Li', 'Li'): G_{Li-Li}(r),
                ('Li', 'Zn'): G_{Li-Zn}(r),
                ('Zn', 'Zn'): G_{Zn-Zn}(r)
            }


        """
        # initalise dict of element pairs with correct keys
        style = self.kwargs.get('style')
        gw = self.kwargs.get('gaussian_width')
        self.r_space = np.arange(0, self.rmax + self.dr, self.dr)
        elem_gr = dict()
        for comb in combinations_with_replacement(set(self._types), 2):
            elem_gr[tuple(set(comb))] = np.zeros_like(self.r_space)

        for elem_type in elem_gr:
            poscart = [self._poscart[i] for i in range(len(self._poscart)) if self._types[i] == elem_type[0]]
            poscart_b = ([self._poscart[i] for i in range(len(self._poscart)) if self._types[i] == elem_type[1]]
                         if len(elem_type) == 2 else None)
            distances = self._calc_distances(poscart, poscart_b=poscart_b)
            elem_gr[elem_type] = (len(elem_type) *
                                  self._get_broadened_normalised_pdf(distances,
                                                                     style=style,
                                                                     gaussian_width=gw))

        self.elem_gr = elem_gr

    def _calc_unprojected_pdf_from_projected(self):
        """" Reconstruct full PDF from projected. """
        self.gr = np.zeros_like(self.r_space)
        for key in self.elem_gr:
            self.gr += self.elem_gr[key]

    @staticmethod
    @numba.njit
    def _normalize_gr(gr, r_space, dr, num_atoms, number_density):
        """ Normalise a broadened PDF, ignoring the Gaussian magnitude. """
        norm = 4 * np.pi * (r_space + dr)**2 * dr * num_atoms * number_density
        return np.divide(gr, norm)

    @staticmethod
    @numba.njit
    def _dist_hist(distances, r_space, dr):
        """ Bin the pair-wise distances according to the radial grid.

        Parameters:
            distances (numpy.ndarray): array of pair-wise distances.
            r_space (numpy.ndarray): radial grid
            dr (float): bin width.

        """
        hist = np.zeros_like(r_space)
        for dij in distances:
            hist[ceil(dij / dr)] += 1
        return hist

    def _get_broadened_normalised_pdf(self, distances, style='smear', gaussian_width=0.1):
        """ Broaden the values provided as distances and return
        G(r) and r_space of the normalised PDF.

        Parameters:
            distances (numpy.ndarray): distances used to calculate PDF

        Keyword arguments:
            style (str): either 'smear' or 'histogram'
            gaussian_width (float): smearing width in Angstrom^1/2

        Returns:
            gr (np.ndarray): G(r), the PDF of supplied distances

        """
        if style == 'histogram' or gaussian_width == 0:
            gr = self._dist_hist(distances, self.r_space, self.dr)
        else:
            # otherwise do normal smearing
            hist = self._dist_hist(distances, self.r_space, self.dr)
            if self.kwargs.get('low_mem'):
                if self.kwargs.get('debug'):
                    print('Using low memory mode...')
                gr = self._broadening_unrolled(hist, self.r_space, gaussian_width)
            else:
                try:
                    gr = self._broadening_distance_dominated(hist, self.r_space, gaussian_width)
                except MemoryError:
                    # if run out of memory, use low memory mode
                    if self.kwargs.get('debug'):
                        print('Ran out of memory, using low memory mode...')
                    self.kwargs['low_mem'] = True
                    gr = self._broadening_unrolled(hist, self.r_space, gaussian_width)

        gr = self._normalize_gr(gr, self.r_space, self.dr, self._num_atoms, self.number_density)

        return gr

    @staticmethod
    def _get_image_trans_vectors_auto(lattice, rmax, dr, max_num_images=50):
        """ Finds all "images" (integer 3-tuples, supercells) that have
        atoms within rmax + dr + longest LV of the parent lattice.

        Parameters:
            lattice (list): list of lattice vectors.
            rmax (float): maximum radial distance.
            dr (float): PDF bin width.

        Keyword arguments:
            max_num_images (int): the greatest integer multiple of LVs to search out to.

        Returns:
            list: list of int 3-tuples of cells, up to rmax or if max_num_images is exceeded,
                just up to 1 cell away.

        """
        image_vec = set()
        # find longest combination of single LV's
        max_trans = 0
        _lattice = np.asarray(lattice)
        products = list(itertools.product(range(-1, 2), repeat=3))
        for prod in products:
            trans = prod @ _lattice
            length = np.sqrt(np.sum(trans**2))
            if length > max_trans:
                max_trans = length

        unit_vector_lengths = np.sqrt(np.sum(_lattice**2, axis=1))
        limits = [int((dr + rmax + max_trans) / length) for length in unit_vector_lengths]

        for ind, limit in enumerate(limits):
            if abs(limit) > max_num_images:
                limits[ind] = int(copysign(max_num_images, limit))

        products = itertools.product(*(range(-lim, lim+1) for lim in limits))
        for prod in products:
            trans = prod @ _lattice
            length = np.sqrt(np.sum(trans**2))
            if length <= rmax + dr + max_trans:
                image_vec.add(prod)

        return image_vec

    def _set_image_trans_vectors(self):
        """ Sets self._image_vec to a list/generator of image translation vectors,
        based on self._num_images.

        If self._num_images is an integer, create all 3-member integer combinations
        up to the value.

        If self._num_images is 'auto', create all translation vectors up to length self.rmax.

        e.g. self._image_vec = [[1, 0, 1], [0, 1, 1], [1, 1, 1]].

        """
        if self._num_images == 'auto':
            self._image_vec = self._get_image_trans_vectors_auto(
                self._lattice, self.rmax, self.dr, max_num_images=self.kwargs.get('max_num_images')
            )
        else:
            self._image_vec = list(itertools.product(range(-self._num_images, self._num_images + 1), repeat=3))

    def get_sim_distance(self, pdf_b, projected=False):
        """ Return the similarity between two PDFs. """
        return PDFOverlap(self, pdf_b, projected=projected).similarity_distance

    def pdf(self):
        """ Return G(r) and the r_space for easy plotting. """
        try:
            return (self.r_space, self.gr)
        except AttributeError:
            return (None, None)

    def plot_projected_pdf(self, **kwargs):
        """ Plot projected PDFs.

        Keyword arguments:
            keys (list): plot only a subset of projections, e.g. [('K', )].
            other_pdfs (list of PDF): other PDFs to plot.

        """
        from matador.plotting.pdf_plotting import plot_projected_pdf
        plot_projected_pdf(self, **kwargs)

    def plot_pdf(self, **kwargs):
        """ Plot PDFs.

        Keyword arguments:
            other_pdfs (list of PDF): other PDFs to add to the plot.

        """
        from matador.plotting.pdf_plotting import plot_pdf
        plot_pdf(self, **kwargs)


class PDFFactory(FingerprintFactory):
    """ This class computes PDF objects from a list of structures,
    as concurrently as possible. The PDFs are stored under the `pdf`
    key inside each structure dict.

    Attributes:
        nprocs (int): number of concurrent processes.

    """
    default_key = 'pdf'
    fingerprint = PDF


class PDFOverlap:
    """ Calculate the PDFOverlap between two PDF objects,
    pdf_a and pdf_b, with number density rescaling.

    Attributes:
        pdf_a (PDF): first PDF to compare.
        pdf_b (PDF): second PDF to compare.
        fine_dr (float): fine grid scale on which to compare.
        similarity_distance (float): "distance" between PDFs.
        overlap_int (float): the value of the overlap integral.

    """

    def __init__(self, pdf_a, pdf_b, projected=False):
        """ Perform the overlap and similarity distance calculations.

        Parameters:
            pdf_a (PDF): first PDF to compare.
            pdf_b (PDF): second PDF to compare.

        Keyword arguments:
            projected : if True, attempt to use projected PDFs.

        """
        self.pdf_a = pdf_a
        self.pdf_b = pdf_b
        self.fine_dr = self.pdf_a.dr / 2.0
        # initialise with large number
        self.similarity_distance = 1e10
        self.overlap_int = 0
        if projected:
            if isinstance(pdf_a.elem_gr, dict) and isinstance(pdf_b.elem_gr, dict):
                self.projected_pdf_overlap()
            else:
                print('Projected PDFs missing, continuing with total.')
            self.pdf_overlap()
        else:
            self.pdf_overlap()

    def pdf_overlap(self):
        """ Calculate the overlap of two PDFs via
        a simple meshed sum of their difference.

        """
        self.overlap_int = 0
        self.similarity_distance = 1e10
        self.fine_space = np.arange(0, self.pdf_a.rmax, self.fine_dr)
        self.fine_gr_a = np.interp(self.fine_space, self.pdf_a.r_space, self.pdf_a.gr)
        self.fine_gr_b = np.interp(self.fine_space, self.pdf_b.r_space, self.pdf_b.gr)
        # scaling factor here is normalising to number density
        density_rescaling_factor = pow(self.pdf_b.number_density / (self.pdf_a.number_density), 1 / 3)
        rescale_factor = density_rescaling_factor
        self.fine_gr_a = np.interp(self.fine_space, rescale_factor * self.fine_space, self.fine_gr_a)
        self.fine_gr_a = self.fine_gr_a[:int(len(self.fine_space) * 0.75)]
        self.fine_gr_b = self.fine_gr_b[:int(len(self.fine_space) * 0.75)]
        self.fine_space = self.fine_space[:int(len(self.fine_space) * 0.75)]
        overlap_fn = self.fine_gr_a - self.fine_gr_b
        worst_case_overlap_int = np.trapz(np.abs(self.fine_gr_a), dx=self.pdf_a.dr/2.0) + \
            np.trapz(np.abs(self.fine_gr_b), dx=self.pdf_b.dr/2.0)
        self.overlap_int = np.trapz(np.abs(overlap_fn), dx=self.pdf_a.dr / 2.0)
        self.similarity_distance = self.overlap_int / worst_case_overlap_int
        self.overlap_fn = overlap_fn

    def projected_pdf_overlap(self):
        """ Calculate the overlap of two projected PDFs via
        a simple meshed sum of their difference.

        """
        self.fine_space = np.arange(0, self.pdf_a.rmax, self.fine_dr)
        self.overlap_int = 0
        self.similarity_distance = 1e10
        elems = set(key for key in self.pdf_a.elem_gr)
        if elems != set(key for key in self.pdf_b.elem_gr):
            for key in self.pdf_b.elem_gr:
                elems.add(key)
        # pad out missing elements with zero PDFs
        for key in elems:
            if key not in self.pdf_a.elem_gr:
                self.pdf_a.elem_gr[key] = np.zeros_like(self.pdf_a.r_space)
            if key not in self.pdf_b.elem_gr:
                self.pdf_b.elem_gr[key] = np.zeros_like(self.pdf_b.r_space)
        self.fine_elem_gr_a, self.fine_elem_gr_b = dict(), dict()
        for key in elems:
            self.fine_elem_gr_a[key] = np.interp(self.fine_space, self.pdf_a.r_space, self.pdf_a.elem_gr[key])
            self.fine_elem_gr_b[key] = np.interp(self.fine_space, self.pdf_b.r_space, self.pdf_b.elem_gr[key])
        # scaling factor here is normalising to number density
        density_rescaling_factor = pow((self.pdf_b.number_density) / (self.pdf_a.number_density), 1 / 3)
        rescale_factor = density_rescaling_factor
        for key in elems:
            self.fine_elem_gr_a[key] = (
                np.interp(self.fine_space, rescale_factor * self.fine_space, self.fine_elem_gr_a[key])
            )
        for key in elems:
            self.fine_elem_gr_a[key] = self.fine_elem_gr_a[key][:int(len(self.fine_space) * 0.75)]
            self.fine_elem_gr_b[key] = self.fine_elem_gr_b[key][:int(len(self.fine_space) * 0.75)]
        self.fine_space = self.fine_space[:int(len(self.fine_space) * 0.75)]

        for key in elems:
            overlap_fn = self.fine_elem_gr_a[key] - self.fine_elem_gr_b[key]
            worst_case_a = np.trapz(np.abs(self.fine_elem_gr_a[key]), dx=self.pdf_a.dr / 2.0)
            worst_case_b = np.trapz(np.abs(self.fine_elem_gr_b[key]), dx=self.pdf_b.dr / 2.0)
            worst_case_overlap = worst_case_a + worst_case_b
            overlap = np.trapz(np.abs(overlap_fn), dx=self.pdf_a.dr / 2.0)
            self.overlap_int += overlap / worst_case_overlap

        self.similarity_distance = self.overlap_int / len(elems)

    def plot_diff(self):
        """ Simple plot for comparing two PDFs. """
        from matador.plotting.pdf_plotting import plot_diff_overlap
        plot_diff_overlap(self)

    def plot_projected_diff(self):
        """ Simple plot for comparing two projected PDFs. """
        from matador.plotting.pdf_plotting import plot_projected_diff_overlap
        plot_projected_diff_overlap(self)


class CombinedProjectedPDF:
    """ Take some computed PDFs and add them together. """
    def __init__(self, pdf_cursor):
        """ Create CombinedPDF object from list of PDFs.

        Parameters:
            pdf_cursor (:obj:`list` of :obj:`PDF`): list of
                PDF objects to combine.

        """
        self.dr = min([pdf.dr for pdf in pdf_cursor])
        self.rmax = min([pdf.rmax for pdf in pdf_cursor])
        self.r_space = np.arange(0, self.rmax + self.dr, self.dr)
        self.label = 'Combined PDF'
        if any([not pdf.elem_gr for pdf in pdf_cursor]):
            raise RuntimeError('Projected PDFs not found.')

        keys = {key for pdf in pdf_cursor for key in pdf.elem_gr}
        self.elem_gr = {key: np.zeros_like(self.r_space) for key in keys}
        for pdf in pdf_cursor:
            for key in pdf.elem_gr:
                self.elem_gr[key] += np.interp(self.r_space, pdf.r_space, pdf.elem_gr[key])

    def plot_projected_pdf(self):
        """ Plot the combined PDF. """
        from matador.plotting.pdf_plotting import plot_projected_pdf
        plot_projected_pdf(self)



# coding: utf-8
# Distributed under the terms of the MIT License.

""" This script mimics the dispersion.pl script bundled
with CASTEP. For the given bands file, a bandstructure
is created. If a <seed>.adaptive.dat file exists
then a combined BS/DOS plot will be created.
"""

from os.path import isfile
import argparse
import glob
from matador import __version__, script_epilog


def main():
    """ Parse args and run the script. """
    parser = argparse.ArgumentParser(
        prog='dispersion',
        epilog=script_epilog,
        description='simple plotting script for bandstructures/DOS based on matador',
        add_help=True)

    parser.add_argument('--version', action='version', version='matador version ' + __version__ + '.')
    parser.add_argument('--pdf', action='store_true',
                        help='save pdf rather than showing plot in X')
    parser.add_argument('--png', action='store_true',
                        help='save png rather than showing plot in X')
    parser.add_argument('--svg', action='store_true',
                        help='save svg rather than showing plot in X')
    parser.add_argument('--labels', type=str, nargs='*',
                        help='list of legend labels, comma separated')
    parser.add_argument('--dos_only', action='store_true', help='only plot DOS')
    parser.add_argument('--bs_only', action='store_true', help='only plot dispersion')
    parser.add_argument('--preserve_kspace_distance', action='store_true',
                        help='when linearising kpoint path, ensure distance in reciprocal space is conserved')
    parser.add_argument('--cmap', type=str,
                        help='matplotlib colourmap name to use')
    parser.add_argument('--n_colours', type=int,
                        help='number of colours to use from colourmap (DEFAULT: 6)')
    parser.add_argument('--spin_only', type=str,
                        help='either "up" or "down" to only plot one channel')
    parser.add_argument('-interp', '--pdis_interpolation_factor', type=float,
                        help='multiple by which to interpolate pDIS bands (DEFAULT: 2)')
    parser.add_argument('-scale', '--pdis_point_scale', type=float,
                        help='point scale in pDIS plots (DEFAULT: 25)')
    parser.add_argument('-p', '--projectors_to_plot', type=str,
                        help=("a list of projectors to plot, in the format element:orbital, "
                              "e.g. -p K:s,P:p. If the orbital is omitted, all orbitals will be used for that element."))
    parser.add_argument('--unstacked_pdos', action='store_true',
                        help='plot PDOS as overlap rather than stack')
    parser.add_argument('--no_pdis', action='store_true',
                        help='do not try to plot PDIS, even if its available')
    parser.add_argument('--band_reorder', action='store_true',
                        help='try to reorder bands based on local gradients')
    parser.add_argument('--pdos_hide_sum', action='store_true',
                        help='plot PDOS without sum pDOS')
    parser.add_argument('--band_colour', type=str,
                        help='override all other colour options with a matploblib-interpretable colour string, '
                             'or choose one of "occ" or "random".')
    parser.add_argument('-g', '--gap', action='store_true',
                        help='plot position and size of band gap')
    parser.add_argument('-ph', '--phonons', action='store_true', default=False,
                        help='plot phonon calculation, rather than electronic')
    parser.add_argument('-ir', '--infrared', action='store_true', default=False,
                        help='plot infrared spectrum from file')
    parser.add_argument('-ir_ss', '--infrared_step_size', type=float, nargs='?', default=1.0,
                        help='step size in cm^{-1} on x-axis for IR plots; must be > 0, default is 1')
    parser.add_argument('-gw', '--gaussian_width', type=float,
                        help=('smearing width for DOS from .bands_dos (default: 0.1 eV) or '
                              '.phonon_dos files (default: 10 1/cm)'))
    parser.add_argument('--highlight_bands', nargs='+', type=int,
                        help='specify band numbres to highlight in plot')
    parser.add_argument('-v', '--verbosity', type=int, default=0,
                        help='control verbosity of output')
    parser.add_argument('-figsize', '--figsize', nargs='+', type=int,
                        help='figure size in inches to pass to matplotlib')
    parser.add_argument('-pw', '--plot_window', nargs='+', type=float,
                        help='energy window [x, y] or [-x, x] to plot either side of E_F (eV)\
                             (DEFAULT: 5 eV)')
    parser.add_argument('seed', type=str, nargs='+',
                        help='seedname or related filename (e.g. bands or dos file)')
    kwargs = vars(parser.parse_args())
    if 'gap' in kwargs:
        gap = kwargs['gap']
        del kwargs['gap']

    seeds = kwargs.get('seed')
    exts_to_strip = ['bands', 'linear.dat', 'adaptive.dat', 'pdis.dat',
                     'bands_dos', 'pdos.dat', 'phonon', 'phonon_dos']

    for ind, seed in enumerate(seeds):
        for ext in exts_to_strip:
            seeds[ind] = seeds[ind].replace('.' + ext, '')

    verbosity = kwargs.get('verbosity')
    phonons = kwargs.get('phonons')
    ir = kwargs.get('infrared')
    if kwargs.get('labels'):
        labels = ' '.join(kwargs.get('labels')).split(',')
    else:
        labels = None

    if kwargs['no_pdis']:
        plot_pdis = False
    else:
        plot_pdis = True
    del kwargs['no_pdis']

    cmap = kwargs.get('cmap')
    band_colour = kwargs.get('band_colour', 'occ')
    band_reorder = kwargs.get('band_reorder', False)

    from matador.plotting import plot_spectral, plot_ir_spectrum

    del kwargs['seed']
    del kwargs['verbosity']
    del kwargs['labels']
    del kwargs['cmap']
    del kwargs['band_colour']
    del kwargs['band_reorder']

    bandstructure = False
    dos = False

    for seed in seeds:
        if not phonons and not ir:
            bs_seed = seed + '.bands'
            bandstructure = isfile(bs_seed)
            dos_seeds = glob.glob(seed + '*.dat')
            if isfile(seed + '.bands_dos'):
                dos_seeds.append(seed + '.bands_dos')
            dos = any([isfile(dos_seed) for dos_seed in dos_seeds])

        elif phonons and not ir:
            phonon_seed = seed + '.phonon'
            bandstructure = isfile(phonon_seed)
            dos_seed = seed + '.phonon_dos'
            dos = isfile(dos_seed)

        elif ir:
            if len(seeds) > 1:
                exit('Multiple seeds not supported for IR plot.')
            ir_seed = seed + '.phonon'

    if bandstructure:
        cell_seed = seed + '.cell'
        cell = isfile(cell_seed)
    else:
        cell = False

    if kwargs.get('dos_only') and dos:
        bandstructure = False

    if kwargs.get('bs_only') and bandstructure:
        dos = False

    if not any([bandstructure, dos, ir]):
        raise SystemExit("Unable to find files for seedname {}".format(seed))

    if not ir:
        return plot_spectral(
            seeds,
            plot_bandstructure=bandstructure,
            plot_dos=dos,
            plot_pdis=plot_pdis,
            cell=cell,
            gap=gap,
            verbosity=verbosity,
            labels=labels,
            cmap=cmap,
            band_reorder=band_reorder,
            band_colour=band_colour,
            **kwargs
        )

    if ir:
        return plot_ir_spectrum(
            ir_seed,
            bin_width=kwargs['infrared_step_size'],
            **kwargs
        )

    exit("Issue plotting {}: did you specify -ph/-ir appropriately?"
         .format(seeds))


if __name__ == '__main__':
    main()



#!/usr/bin/env python
import unittest
import os
import shutil
from matador.config import load_custom_settings, set_settings
from matador.config.config import DEFAULT_SETTINGS

DUMMY_SETTINGS = {
    "mongo": {"host": "blah", "port": 666},
    "plotting": {"style": "matador"},
    "this_is_a_test": {True: "it is"},
    "run3": {
        "castep_executable": "castep",
        "optados_executable": "optados",
        "scratch_prefix": ".",
    },
}

REAL_PATH = "/".join(__file__.split("/")[:-1]) + "/"


class ConfigTest(unittest.TestCase):
    """ Test config loading. """

    def tearDown(self):
        from matador.config import SETTINGS

        SETTINGS.reset()

    def setUp(self):
        from matador.config import SETTINGS

        SETTINGS.reset()

    def testLoadNamedCustomSettings(self):
        """ Test custom config. """
        settings = load_custom_settings(
            config_fname=(REAL_PATH + "data/custom_config.yml"), no_quickstart=True
        )
        self.assertEqual(settings.settings, DUMMY_SETTINGS)
        from matador.config import SETTINGS

        self.assertEqual(SETTINGS.settings, DUMMY_SETTINGS)
        SETTINGS.reset()

    def testSetSettings(self):
        set_settings(DUMMY_SETTINGS)
        from matador.config import SETTINGS

        self.assertEqual(SETTINGS.settings, DUMMY_SETTINGS)

        SETTINGS["backend"] = "mongo"
        self.assertEqual(SETTINGS.settings["backend"], "mongo")
        SETTINGS.reset()

    def testLoadUserDefaultSettings(self):
        """ Test default config. """
        exists = False
        try:
            if os.path.isfile(os.path.expanduser("~/.matadorrc")):
                exists = True
                shutil.copy(
                    os.path.expanduser("~/.matadorrc"),
                    os.path.expanduser("~/.matadorrc_bak"),
                )
            shutil.copy(
                REAL_PATH + "data/custom_config.yml", os.path.expanduser("~/.matadorrc")
            )
            settings = load_custom_settings(no_quickstart=True)
            self.assertEqual(settings.settings, DUMMY_SETTINGS)
            os.remove(os.path.expanduser("~/.matadorrc"))
            if exists:
                shutil.copy(
                    os.path.expanduser("~/.matadorrc_bak"),
                    os.path.expanduser("~/.matadorrc"),
                )
                os.remove(os.path.expanduser("~/.matadorrc_bak"))
        except Exception as oops:
            if exists:
                shutil.copy(
                    os.path.expanduser("~/.matadorrc_bak"),
                    os.path.expanduser("~/.matadorrc"),
                )
                os.remove(os.path.expanduser("~/.matadorrc_bak"))
            raise oops

    def testLoadDefaultSettings(self):
        """ Test default config. """
        settings = load_custom_settings(
            config_fname="definitely_doesnt_exist.yml", no_quickstart=True
        )
        self.assertEqual(settings.settings, DEFAULT_SETTINGS)

        settings.reset()
